import os
import shutil
import click
import git
import urnc.util as util
import urnc.logger as log
import dateutil.parser

from urnc.convert import convert_fn
from datetime import datetime

def clone_student_repo(config):
    url = util.get_config_value(
        config, "git", "student", required=True)
    assert (url is not None)
    repo_name_git = os.path.basename(url)
    (repo_name, _) = os.path.splitext(repo_name_git)
    folder_name = f"{repo_name}-student"

    base_path = os.path.dirname(os.getcwd())
    repo_path = os.path.join(base_path, folder_name)

    if os.path.exists(repo_path):
        try:
            repo = git.Repo(repo_path)
            repo_url = repo.remote().url
            if (repo_url != url):
                return log.critical(f"Repo remote mismatch {repo_url}!={url}")
            log.log(f"Returning existing repo {repo_path}")
            return repo
        except Exception:
            return log.critical(f"{repo_path} exists but is not a git repo")
    log.log(f"Cloning student repo {url} to {repo_path}")
    repo = git.Repo.clone_from(url, repo_path)
    util.update_repo_config(repo)
    return repo


def clear_repo(repo):
    path = repo.working_dir
    entries = os.listdir(path)
    for entry in entries:
        if (entry.startswith(".")):
            continue
        entry_path = os.path.join(path, entry)
        if os.path.isfile(entry_path):
            os.remove(entry_path)
        if os.path.isdir(entry_path):
            shutil.rmtree(entry_path)


def copy_files(repo, student_repo):
    path = repo.working_dir
    target_path = student_repo.working_dir
    entries = os.listdir(path)
    for entry in entries:
        if (entry.startswith(".")):
            continue
        entry_path = os.path.join(path, entry)
        copy_path = os.path.join(target_path, entry)
        if os.path.isfile(entry_path):
            shutil.copy2(entry_path, copy_path)
        if os.path.isdir(entry_path):
            shutil.copytree(entry_path, copy_path)


def write_gitignore(repo, student_repo, config):
    main_gitignore = os.path.join(repo.working_dir, ".gitignore")
    student_gitignore = os.path.join(student_repo.working_dir, ".gitignore")
    if (os.path.exists(main_gitignore)):
        shutil.copy(main_gitignore, student_gitignore)
    exclude = util.get_config_value(config, "git", "exclude", default=[])
    now = datetime.now()
    with open(student_gitignore, "a") as gitignore:
        for value in exclude:
            if isinstance(value, str):
                gitignore.write(f"{value}\n")
            else:
                if "after" in value:
                    if now < dateutil.parser.parse(value["after"]):
                        continue
                if "until" in value:
                    if now > dateutil.parser.parse(value["until"]):
                        continue
                gitignore.write(f"{value['pattern']}\n")


def update_index(repo):
    cached_files_str = repo.git.ls_files("-ci", "--exclude-standard")
    if (cached_files_str != ''):
        cached_files = cached_files_str.split("\n")
        log.log(f"Removing excluded files {cached_files}")
        repo.index.remove(cached_files, working_tree=False)
        repo.index.write()


@ click.command(short_help="Build student version and push to public repo", help="Run the urnc ci pipeline, this pushes the converted notebooks to the public repo")
@ click.pass_context
def ci(ctx):
    ci_fn(ctx, True)


def ci_fn(ctx, commit=True):
    log.setup_logger(use_file=False)
    config = util.read_config(ctx)
    repo = util.get_git_repo(ctx)
    if (repo.is_dirty() and commit):
        raise Exception(f"Repo is not clean. Commit your changes.")

    student_repo = clone_student_repo(config)
    clear_repo(student_repo)
    copy_files(repo, student_repo)

    log.log("Converting files")
    convert_fn(ctx,
               input=repo.working_dir,
               output=student_repo.working_dir,
               force=True,
               dry_run=False
               )
    log.log("Notebooks converted")

    log.log("Updating .gitignore from config")
    write_gitignore(repo, student_repo, config)

    # Remove exclude files from active index
    log.log("Dropping cached files...")
    update_index(student_repo)

    if not commit:
        log.log("Skipping git commit and push")
        log.log("Done.")
        return

    log.log("Adding files and commiting")
    student_repo.git.add(all=True)
    student_repo.index.commit("urnc convert")
    log.log("Pushing student repo")
    student_repo.git.push()
    log.log("Done.")



from os.path import abspath
import click
import nbformat
import os
from traitlets.config import Config
from nbconvert.preprocessors.clearoutput import ClearOutputPreprocessor
from nbconvert.exporters.notebook import NotebookExporter

from urnc.preprocessor.add_tags import AddTags
from urnc.preprocessor.assignments import ProcessAssignments
from urnc.preprocessor.remove_solutions import RemoveSolutions
from urnc.preprocessor.broken_links import BrokenLinks

import urnc.logger as log


@click.command(help="Convert notebooks")
@click.argument(
    "input",
    type=click.Path(exists=True),
    default="."
)
@click.argument(
    "output",
    type=str,
    default="out"
)
@click.option("-v", "--verbose", is_flag=True)
@click.option("-f", "--force", is_flag=True)
@click.option("-n", "--dry-run", is_flag=True)
@click.pass_context
def convert(ctx, input, output, verbose, force, dry_run):
    log.setup_logger(False, verbose)
    convert_fn(ctx, input, output, force, dry_run)


@click.command(help="Check notebooks for errors")
@click.argument(
    "input",
    type=click.Path(exists=True),
    default="."
)
@click.pass_context
@click.option("-q", "--quiet", is_flag=True)
def check(ctx, input, quiet):
    log.setup_logger(False, not quiet)
    convert_fn(ctx, input, None, False, True)


def get_abs_path(ctx, path):
    if (path is None):
        return None
    if (os.path.isabs(path)):
        return path
    base: str = ctx.obj["ROOT"]
    new_path = os.path.join(base, path)
    return os.path.abspath(new_path)


def convert_fn(ctx, input, output, force, dry_run):
    input = get_abs_path(ctx, input)
    if (input is None):
        raise Exception("No input")
    output = get_abs_path(ctx, output)
    log.log(f"Converting notebooks in {input} to {output}")
    paths = []
    folder = input
    if os.path.isfile(input):
        paths.append(input)
        folder = os.path.dirname(input)
    else:
        for root, dirs, files in os.walk(input, topdown=True):
            dirs[:] = [d for d in dirs if not d[0] == "."]
            for file in files:
                if file.lower().endswith(".ipynb"):
                    paths.append(os.path.join(root, file))

    if len(paths) == 0:
        log.warn("No notebooks found in input %s" % input)
        return

    c = Config()
    c.NotebookExporter.preprocessors = [
        BrokenLinks,
        AddTags,
        ProcessAssignments,
        RemoveSolutions,
        ClearOutputPreprocessor,
    ]
    converter = NotebookExporter(config=c)

    for file in paths:
        opath = os.path.relpath(file, folder)
        out_file = None
        out_dir = None
        if output is not None:
            out_file = os.path.join(output, opath)
            out_dir = os.path.dirname(out_file)
            os.makedirs(out_dir, exist_ok=True)
        if out_file is None:
            log.log(f"Checking {file}")
        elif os.path.isfile(out_file):
            if not force:
                log.log(f"Skipping {file} because {out_file} already exists")
                continue
            else:
                log.warn(f"Overwriting {out_file}")
        else:
            log.log(f"Converting '{file}' into '{out_file}'")
        notebook = nbformat.read(file, as_version=4)
        resources = {}
        resources["path"] = file
        (output_text, _) = converter.from_notebook_node(notebook, resources)

        if (out_file is None or dry_run):
            continue
        with open(out_file, "w") as f:
            f.write(output_text)



from nbconvert.preprocessors.base import Preprocessor

import os
import re
import requests
import urnc.preprocessor.util as util
import urnc.logger as log
from urnc.preprocessor.util import Keywords, Tags


def url_is_valid(url):
    try:
        response = requests.head(url, allow_redirects=True)
        return response.status_code == 200
    except Exception:
        return False


def check_image(base_folder, src):
    if src.startswith("http"):
        log.warn(f"Remote image detected. {src}")
        if not url_is_valid(src):
            log.error(f"Request to {src} failed.")
        return
    image_path = os.path.normpath(os.path.join(base_folder, src))
    if (not os.path.exists(image_path)):
        log.warn(f"The image {image_path} does not exists.")


class BrokenLinks(Preprocessor):
    def preprocess(self, notebook, resources):
        path = resources["path"]

        base_folder = os.path.dirname(path)

        for cell in notebook.cells:
            matches = re.findall(Keywords.IMAGE_TAG, cell.source)
            for match in matches:
                log.dbg(f"Detected html image tag with src {match}")
                check_image(base_folder, match)

            matches = re.findall(Keywords.MD_IMAGE_TAG, cell.source)
            for match in matches:
                (alt, src) = match
                log.dbg(f"Detected markdown image tag with src {src}")
                check_image(base_folder, src)

        return notebook, resources



import inspect
import re
from nbconvert.preprocessors.base import Preprocessor

import urnc.preprocessor.util as util
from urnc.preprocessor.util import Keywords

import urnc.logger as log


def replace_local_link(cell):
    if cell.cell_type != "markdown":
        return
    if not re.search(Keywords.ASSIGNMENT_LINK, cell.source, re.IGNORECASE):
        return
    log.dbg("Detected Link to assignment.")
    cell.source = re.sub(Keywords.ASSIGNMENT_REPLACE,
                         "#a-", cell.source, re.IGNORECASE)


class ProcessAssignments(Preprocessor):
    def preprocess(self, notebook, resources):
        for cell in notebook.cells:
            replace_local_link(cell)
            if util.has_tag(cell, util.Tags.ASSIGNMENT_START):
                eid = cell.metadata.assignment_id
                html = f"""
                    <a class="anchor" name="a-{eid}"></a>
                    <h3>   ✎    Assignment {eid}</h3>
                """
                cell.source = inspect.cleandoc(html)
        return notebook, resources



import click
import os
import git


import urnc.util as util

import urnc.logger as log


def get_upstream_changes(repo):
    branch_name = repo.active_branch
    changes = repo.git.diff(
        f"..origin/{branch_name}", "--name-status").split('\n')
    return changes


def get_upstream_deleted(repo):
    changes = get_upstream_changes(repo)
    files = []
    for change in changes:
        parts = change.split('\t', 1)
        if (len(parts) != 2):
            continue
        [ctype, file] = change.split('\t', 1)
        if (ctype == "D"):
            files.append(file)
    return files


def get_upstream_added(repo):
    changes = get_upstream_changes(repo)
    files = []
    for change in changes:
        parts = change.split('\t', 1)
        if (len(parts) != 2):
            continue
        [ctype, file] = parts
        if (ctype == "A"):
            files.append(file)
    return files


def get_course_info(course_name, output):
    base_url = os.getenv("GIT_URL")
    if not base_url:
        log.error("Failed to pull. Missing env var GIT_URL")
        return (None, None)

    if not course_name:
        try:
            git_repo = git.Repo(os.getcwd(), search_parent_directories=True)
            git_url = git_repo.remote().url
            if not git_url.startswith(base_url):
                log.error(f"Failed to pull. Invalid git remote_url {git_url}")
                return (None, None)
            folder_name = git_repo.working_dir
            return (git_url, folder_name)
        except Exception:
            log.error("Failed to pull no course_name given and not in a git repo")
            return (None, None)
    folder_name = output
    if not output:
        folder_name = course_name

    git_url = f"{base_url}{course_name}.git"
    return (git_url, folder_name)


def get_repo(folder_name):
    base_url = os.getenv("GIT_URL")
    if not base_url:
        log.error("Missing env var GIT_URL")
        return None
    try:
        git_repo = git.Repo(folder_name)
        git_url = git_repo.remote().url
        if not git_url.startswith(base_url):
            log.error(
                f"Invalid git remote_url {git_url} needs to start with {base_url}")
            return None
        return git_repo
    except Exception:
        log.error(f"{folder_name} is not a git repo")
        return None


def reset_deleted_files(repo):
    branch = repo.active_branch
    deleted_files = repo.git.ls_files("--deleted").split('\n')
    deleted_upstream = get_upstream_deleted(repo)
    for filename in deleted_files:
        if (not filename):
            continue
        if (filename in deleted_upstream):
            repo.git.checkout("HEAD", '--', filename)
        else:
            log.dbg(f"Restoring deleted file {filename}")
            repo.git.checkout(f"origin/{branch}", "--", filename)


def merge(repo):
    branch = repo.active_branch
    remote_branch = f"origin/{branch}"
    repo.git.merge("-Xours", remote_branch)


def pull_admin(course_name, output, branch, depth):
    isAdmin = os.getenv('IS_ADMIN') == "1"
    if (not isAdmin):
        return
    if (not course_name):
        return
    base_url = os.getenv("GIT_URL_ADMIN")
    if (not base_url):
        log.error("Missing env var GIT_URL_ADMIN")
        return
    git_url = f"{base_url}{course_name}.git"
    folder_name = output
    if not output:
        folder_name = f"{course_name}-admin"

    if (not os.path.exists(folder_name)):
        log.log(f"{folder_name} does not exists. Cloning repo {git_url}")
        try:
            git.Repo.clone_from(git_url, folder_name,
                                branch=branch, depth=depth)
            log.log("Cloned successfully.")
        except Exception as err:
            log.error("Failed to clone repo. Error:")
            log.error(err)
    else:
        log.log(f"{folder_name} exists.")
        try:
            repo = git.Repo(folder_name)
            log.log(f"Pulling...")
            repo.git.pull("--ff-only")
            log.log(f"Done")

        except Exception as err:
            log.error(f"{folder_name} is not a git repo")


@click.command(help="Pull the course repo")
@click.argument(
    "course_name",
    type=str,
    default=None,
    required=False
)
@click.option("-o", "--output", type=str, help="The name of the output folder", default=None)
@click.option("-b", "--branch", help="The branch to pull", default="main")
@click.option("-d", "--depth", help="The depth for git fetch", default=1)
@click.pass_context
def pull(ctx, course_name, output, branch, depth):
    log.setup_logger()
    try:
        pull_fn(ctx, course_name, output, branch, depth)
    except Exception as err:
        log.error("pull failed with unexpected error.")
        log.error(err)


def pull_fn(ctx, course_name, output, branch, depth):
    pull_admin(course_name, output, branch, depth)

    (git_url, folder_name) = get_course_info(course_name, output)
    if (not git_url or not folder_name):
        return

    if (not os.path.exists(folder_name)):
        log.log(f"{folder_name} does not exists. Cloning repo {git_url}")
        try:
            git.Repo.clone_from(git_url, folder_name,
                                branch=branch, depth=depth)
            log.log("Cloned successfully.")
        except Exception as err:
            log.error("Failed to clone repo. Error:")
            log.error(err)
        return
    log.log(f"{folder_name} exists. Updating the repo")
    repo = get_repo(folder_name)
    if (not repo):
        return

    log.log(f"Fetching changes...")
    repo.remote().fetch()
    log.log(f"Restoring locally deleted files")
    reset_deleted_files(repo)
    log.log(f"unstaging all changes")
    repo.git.reset("--mixed")

    util.update_repo_config(repo)
    if repo.is_dirty():
        log.log("Repo is dirty. Commiting....")
        repo.git.commit("-am", "Automatic commit by urnc",
                        "--allow-empty")
        log.log("Created new commit")

    log.log("Merging from remote...")
    merge(repo)
    log.log("Done.")



import click
import yaml
import tomllib
import tomli_w
import git
import os


def branch_exists(repo, branch):
    origin_branch = f"origin/{branch}"
    for ref in repo.references:
        if (ref.name == branch or ref.name == origin_branch):
            return True
    return False


def update_repo_config(repo):
    try:
        r = repo.config_reader()
        name = r.get_value("user", "name")
        email = r.get_value("user", "email")
        assert (name is not None)
        assert (email is not None)
        return
    except Exception:
        pass

    config_writer = repo.config_writer()
    config_writer.set_value("user", "name", "urnc")
    config_writer.set_value("user", "email", "urnc@spang-lab.de")
    config_writer.release()


def get_config_value(config, *args, default=None, required=False):
    value = config
    full_key = "config"
    for key in args:
        full_key = f"{full_key}.{key}"
        if key not in value:
            if required:
                raise Exception(f"{full_key} is required")
            return default
        value = value[key]
    return value


def get_git_repo(ctx):
    path = ctx.obj["ROOT"]
    try:
        git_repo = git.Repo(path, search_parent_directories=True)
        return git_repo
    except Exception:
        raise click.UsageError(
            f'The current working directory "{path}" is not a git repo'
        )


def get_git_root(ctx):
    repo = get_git_repo(ctx)
    return repo.working_dir


def read_config(ctx):
    filename = "config.yaml"
    base_path = get_git_root(ctx)

    path = os.path.join(base_path, filename)

    if not os.path.isfile(path):
        raise click.UsageError(
            f"urnc expects a config file called {filename} "
            f"in the git root folder '{base_path}' "
            "make sure you a in a course directory",
        )
    try:
        with open(path, "r") as f:
            config = yaml.safe_load(f)
        if "git" in config and "student" in config["git"]:
            config["git"]["student"] = config["git"]["student"].format(
                **os.environ)
        return config
    except Exception as e:
        raise click.FileError(path, str(e))


def write_config(ctx, data):
    filename = "config.yaml"
    base_path = get_git_root(ctx)

    path = os.path.join(base_path, filename)
    try:
        with open(path, "w") as f:
            yaml.dump(data, f, default_flow_style=False, sort_keys=False)
            return path
    except Exception as e:
        raise click.FileError(path, str(e))


def read_pyproject(ctx):
    filename = "pyproject.toml"
    base_path = get_git_root(ctx)

    path = os.path.join(base_path, filename)

    try:
        with open(path, "rb") as f:
            config = tomllib.load(f)
            return config
    except Exception as e:
        raise click.FileError(path, str(e))


def write_pyproject(ctx, data):
    filename = "pyproject.toml"
    base_path = get_git_root(ctx)

    path = os.path.join(base_path, filename)
    try:
        with open(path, "wb") as f:
            tomli_w.dump(data, f)
            return path
    except Exception as e:
        raise click.FileError(path, str(e))



from nbconvert.preprocessors.base import Preprocessor

import re
import urnc.preprocessor.util as util
from urnc.preprocessor.util import Keywords, Tags
import urnc.logger as log


def end_assignment(assignment_id, has_solution):
    if assignment_id is None:
        return
    if not has_solution:
        log.warn(f"Assignment {assignment_id} has no solution")


class AddTags(Preprocessor):
    def preprocess(self, notebook, resources):
        assignment_ids = set()
        assignment_id = None
        has_solution = False

        for cell in notebook.cells:
            if re.match(Keywords.ASSIGNMENT_DEPRECATED, cell.source, re.IGNORECASE):
                log.warn(
                    "'Exercise' as Keyword is deprecated. Use 'Assignment' instead")
            if match := re.search(Keywords.ASSIGNMENT_START, cell.source, re.IGNORECASE):
                end_assignment(assignment_id, has_solution)
                has_solution = False
                assignment_id = match.group(1)
                util.set_tag(cell, Tags.ASSIGNMENT_START)
                if assignment_id in assignment_ids:
                    log.error("Duplicate Assignment id '%s'" % assignment_id)
                else:
                    assignment_ids.add(assignment_id)
                log.dbg("Detected Assignment '%s'" % assignment_id)
            elif util.has_header(cell) and assignment_id is not None:
                end_assignment(assignment_id, has_solution)
                assignment_id = None
            if assignment_id is None:
                continue
            util.set_tag(cell, Tags.ASSIGNMENT)
            cell.metadata.assignment_id = assignment_id

            if re.search(Keywords.SOLUTION, cell.source, re.IGNORECASE) or util.has_tag(
                cell, Tags.SOLUTION
            ):
                log.dbg(" Detected Solution cell %s" %
                        util.cell_preview(cell))
                util.set_tag(cell, Tags.SOLUTION)
                has_solution = True
        end_assignment(assignment_id, has_solution)
        return notebook, resources



import enum
import re


class Keywords(str, enum.Enum):
    ASSIGNMENT_DEPRECATED = r"^### Exercise `?([\w-]+)`?"
    ASSIGNMENT_START = r"^### (?:Exercise|Assignment) `?([\w-]+)`?"
    SOLUTION = "### Solution"
    SKELETON = "### Skeleton"
    SOLUTION_END = "###"
    IMAGE_TAG = r'<img[^>]*src="([^"]*)"'
    MD_IMAGE_TAG = r"!\[([^\]]*)\]\(([^)]*)\)"
    ASSIGNMENT_LINK = r"\(#Assignment-.*\)"
    ASSIGNMENT_REPLACE = r"#Assignment-"


class Tags(str, enum.Enum):
    SOLUTION = "solution"
    ASSIGNMENT = "assignment"
    ASSIGNMENT_START = "assignment-start"
    SKELETON = "skeleton"


def cell_preview(cell):
    short = cell.source[0:100]
    text = re.sub(r"[\s\n]+", " ", short)
    return "'%s...'" % text[0:50]


def has_tag(cell, tag):
    ltag = tag.lower()
    if "tags" not in cell.metadata:
        return False
    for tag in cell.metadata.tags:
        if tag.lower() == ltag:
            return True
    return False


def set_tag(cell, tag):
    if has_tag(cell, tag):
        return
    if "tags" not in cell.metadata:
        cell.metadata.tags = []
    cell.metadata.tags.append(tag)


def has_header(cell):
    if cell.cell_type != "markdown":
        return False
    if re.search(Keywords.SOLUTION, cell.source):
        return False
    if re.search("^#", cell.source):
        return True
    if re.search("\n#", cell.source):
        return True
    return False
